//=============================================================================
//  @brief  Implementations (bodies) of extern functions for the add RESSL command
//  @author Bryan Morris, Jeff McNeal, Verilab (www.verilab.com)
// =============================================================================
//
//                      RESSL SystemVerilab Utilities Library
//
// @File: ressl_impl_Commands_add.svh
//
// Copyright 2015 Verilab, Inc.
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
//=============================================================================

// -------------------- Strategy / Policy patterns

// ---- add_cmd Implementation

// Syntax: add <seq_name> [<repeat_count>]
//   seq_name = string pointing to a valid sequence name that must exist in the
//      sequence registry.  Otherwise an error message is generated and nothing
//      changes.
//   repeat_count = an optional parameter to allow the user to create multiple
//      copies of the 
                                    
task add_cmd::execute( 
   input string args[$],
   ref  boolean is_command_valid
);
   uvm_sequence_base my_registry_seq;
   uvm_sequence_base my_uvm_seq, my_new_seq;
   seq_info my_seq_info_q[$];
   string       my_seq_name;
   integer      my_repeat_count;

   // ensure that the sequence does NOT exist.
   my_seq_name = args[0]; 
   if (this.ressl_.m_sequence_registry.exists(my_seq_name) == 0) begin
      this.ressl_.stderr( .err_msg( "try again... this time with a sequence name doesn't already exist." ),
                          .help_msg( this.help( ) ));
      is_command_valid = FALSE;
   end else begin
      my_repeat_count = this.get_repeat_count(args[1]);
      if (my_repeat_count == 0) begin
         // Uhh.. that makes no sense.... we've already complained, so just leave.
         is_command_valid = FALSE;
         return;
      end
      
      repeat(my_repeat_count) begin
         // It's unlikely by this point (since we know from above that this seq
         // name exists.. but we use the return value regardless.  Why waste it?
         if (this.ressl_.m_sequence_registry.get(my_seq_name, my_seq_info_q) == 1) begin
            foreach (my_seq_info_q[i]) begin 
               // clone and add each seq in this sub-sequence into the currently selected sequence.   
               my_uvm_seq = my_seq_info_q[i].get_seq();
               my_new_seq = uvm_sequence_base'(my_uvm_seq.clone( )); 
               this.ressl_.add_to_current_sequence( my_new_seq, my_seq_info_q[i].get_seqr() );
            end
         end 
      end
   end
endtask : execute

function string add_cmd::help( );
   help = "add <sequence name> [repeat_count]";
endfunction : help


// Function: get_repeat_count
//    If the user supplies a repeat count (i.e., to add multiple copies, convert it),
//    otherwise just add 1 copy.  If they supply a 0.. that just makes no sense
//    so complain and return 0.
function int unsigned add_cmd::get_repeat_count(
   input string repeat_count_arg
);
   if (repeat_count_arg == "") begin
      get_repeat_count = 1;
   end else begin
      get_repeat_count = repeat_count_arg.atoi();
   end
   
   if (get_repeat_count == 0) begin
      this.ressl_.stderr(.err_msg("You set the repeat_count to 0?!  Really?!"),
                         .help_msg(this.help()));
   end
endfunction : get_repeat_count                    