//=============================================================================
//  @brief  Implementations (bodies) of extern functions for the attach RESSL command
//  @author Bryan Morris, Jeff McNeal, Verilab (www.verilab.com)
// =============================================================================
//
//                      RESSL SystemVerilab Utilities Library
//
// @File: ressl_impl_Commands_attach.svh
//
// Copyright 2015 Verilab, Inc.
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
//=============================================================================

// -------------------- Strategy / Policy patterns

// ---- attach_cmd Implementation

// Syntax: move <src_index> <dst_index>
//   Allows you to move the sub-sequence indicated by <src_index> to <dst_index>
//   i.e., change the order of the sub-sequences. 
task attach_cmd::execute( 
   input string args[$],
   ref  boolean is_command_valid
);
   string       my_seqr_name = args[0];
   string       my_seq_name = args[1];
   string       my_seq_index = args[2];
                
   seq_info     my_seq_info_q[$];
   uvm_object   my_obj;   // temp during creation

   // Does the sequence exist in the registry?
   if ( !this.ressl_.m_sequence_registry.exists( my_seq_name ) ) begin
      // No.  Complain.
      this.ressl_.stderr(.err_msg($sformatf("Sequence specified (%s) does not exist.", 
                                           my_seq_name)),
                        .help_msg("Create the sequence first and then attach it to a sequencer."));
      is_command_valid = FALSE;
      return;
   end
    
   // Check to see if the sequencer exists
   if ( !this.ressl_.m_sequencer_registry.exists( my_seqr_name ) ) begin
      // No.  Complain.
      this.ressl_.stderr(.err_msg($sformatf("Sequencer specified (%s) does not exist.", 
                                           my_seqr_name)),
                        .help_msg("Create the sequencer first and then attach the sequence."));
      is_command_valid = FALSE;
      return;
   end
   // At this point, the sequence AND the sequencer are both known to RESSL.
   // Given that we *know* the seq_name exists, the next check is superfluous.
   if (this.ressl_.m_sequence_registry.get(my_seq_name, my_seq_info_q) == 1) begin

      // OK, we decide whether to attach to ALL sequences in this q or to a specific sequence.
      if ((my_seq_index != "ALL" || my_seq_index == "all")) begin
         int unsigned my_seq_i = my_seq_index.atoi();
         this.attach(.seq_seqr_info(my_seq_info_q[my_seq_i]),
                     .seqr(this.ressl_.m_sequencer_registry[my_seqr_name]));
      end
      else begin
         foreach (my_seq_info_q[i]) begin
            this.attach(.seq_seqr_info(my_seq_info_q[i]),
                        .seqr(this.ressl_.m_sequencer_registry[my_seqr_name]));
         end
      end
   end 
endtask : execute

function void attach_cmd::attach(
   ref   seq_info           seq_seqr_info,
   input uvm_sequencer_base seqr
);
   uvm_sequence_base my_seq;

   my_seq = seq_seqr_info.get_seq();
   // Does the sequence already *have* a sequencer attached?
   if (seq_seqr_info.get_seqr() != null) begin
      // Yes, override it.
      string prev_seqr_name = seq_seqr_info.get_seqr().get_name();

      this.ressl_.stdout($sformatf("Overriding attachment on seq: %s to seqr: %s (was %s)\n",
                                   my_seq.get_name(),
                                   seqr.get_name(),
                                   prev_seqr_name));
   end
   else begin
      this.ressl_.stdout($sformatf("Attaching seq: %s to seqr: %s\n",
                                   my_seq.get_name(),
                                   seqr.get_name()));
   end 
   
   // Now that we told them what we've done, do it. 
   seq_seqr_info.set_seqr(seqr);
   
endfunction : attach

function string attach_cmd::help( );
   help = "attach <seqr> <to_seq> [ALL | <seq_index>]";
endfunction : help


