//=============================================================================
//  @brief  Implementations (bodies) of extern functions RESSL create command (policy pattern)
//  @author Bryan Morris, Jeff McNeal, Verilab (www.verilab.com)
// =============================================================================
//
//                      RESSL SystemVerilab Utilities Library
//
// @File: ressl_impl_Commands.svh
//
// Copyright 2015 Verilab, Inc.
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
//=============================================================================

// -------------------- Strategy / Policy patterns

// ---- create_cmd Implementation
                                    
// Syntax : create <SEQ | SEQR> <type> <name>
task create_cmd::execute( 
   input string args[$],
   ref  boolean is_command_valid
 );
   string       my_seq_or_seqr, my_type, my_name;
   uvm_object   my_obj;   // temp during creation

   // Grab the right sequence
   my_seq_or_seqr = args[0];
   my_type = args[1];
   my_name = args[2];

   // Try and create the supplied type (by name)
   my_obj = this.ressl_.factory_.create_object_by_name(.requested_type_name(my_type),
                                                       .name(my_name));
   if (my_obj == null) begin
      this.ressl_.stderr( .err_msg($sformatf("Failed to find sequence type %s.\n", my_type)),
                          .help_msg("For this command you need to know the exact name of the sequence \nyou created and registered with the uvm_factory.") );
      is_command_valid = FALSE;
   end else begin
      case (my_seq_or_seqr)
        "SEQ", "seq":   this.add_seq(my_name, my_obj, my_type, is_command_valid);
        "SEQR", "seqr": this.add_seqr(my_name, my_obj, my_type, is_command_valid);
        default: begin
           this.ressl_.stderr( .err_msg("Invalid create decl defined.  Only SEQ or SEQR are currently allowed"),
                               .help_msg("Syntax: create <SEQ | SEQR> <type> <name>"));
        end
      endcase // case (my_seq_or_seqr)
   end

   // TODO: automagically call the list cmd
endtask : execute 

//
function void create_cmd::add_seq(
   input string     seq_name,
   input uvm_object seq, 
   input string     seq_type,                         
   ref boolean is_command_valid                      
);
   uvm_sequence_base my_uvm_seq;
   
   // Now attempt to cast it to a sequence type
   if ($cast(my_uvm_seq, seq) == 0) begin
      this.ressl_.stderr( .err_msg($sformatf("Failed to cast type %s to a uvm_sequence type", seq_type)),
                          .help_msg("The typename you supply *really* needs to be derived from a uvm_sequence_base.") );
      is_command_valid = FALSE;
   end

   if (this.ressl_.add_to_sequence_registry(seq_name, my_uvm_seq) == FALSE) begin
      this.ressl_.stderr( .err_msg($sformatf("That's odd... cannot add %s to the sequence registry", seq_name)),
                          .help_msg("...uhh... do you already something in the register with the same name?! Or debug it and send us the fix. ;-)") );
      is_command_valid = FALSE;
   end
endfunction : add_seq

//
function void create_cmd::add_seqr(
   input string     seqr_name, 
   input uvm_object seqr, 
   input string     seqr_type,                                 
   ref boolean is_command_valid                      
);
   uvm_sequencer_base my_uvm_seqr;
   
   // Now attempt to cast it to a sequence type
   if ($cast(my_uvm_seqr, seqr) == 0) begin
      this.ressl_.stderr( .err_msg($sformatf("Failed to cast type %s to a uvm_sequencer type", seqr_type)),
                          .help_msg("The typename you supply *really* needs to be derived from a uvm_sequence_base.") );
      is_command_valid = FALSE;
   end

   if (this.ressl_.add_to_sequencer_registry(seqr_name, my_uvm_seqr) == FALSE) begin
      this.ressl_.stderr( .err_msg($sformatf("That's odd... cannot add %s to the sequencer registry", seqr_name)),
                          .help_msg("...uhh... do you already something in the register with the same name?! Or debug it and send us the fix. ;-)") );
      is_command_valid = FALSE;
   end
endfunction : add_seqr

function string create_cmd::help( );
   help = "create <SEQ | SEQR> <type> <name>";
endfunction : help

