//=============================================================================
//  @brief  Implementations (bodies) of extern functions RESSL objects
//  @author Bryan Morris, Jeff McNeal Verilab (www.verilab.com)
// =============================================================================
//
//                      RESSL SystemVerilab Utilities Library
//
// @File: ressl_impl_Ressl.svh
//
// Copyright 2015 Verilab, Inc.
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
//=============================================================================


// ressl Implementation

function ressl::new (string name="");
   super.new(name);

   // Most importantly, create our sequence registry.
   this.m_sequence_registry = registry#(string, seq_info)::type_id::create("m_sequence_registry");
   
   // Indicate we've not selected *any* sequence.
   this.current_sequence_name_ = "*";
   
   // populate the verb command for the interpreter
   // TODO: Add the ability to "self-append" a ressl_cmd to this table.
   //     Will make the DSL extensisble.
   `add_new_ressl_cmd(this, help)
    
   `add_new_ressl_cmd(this, list )
   `add_new_ressl_cmd(this, describe )

   `add_new_ressl_cmd(this, select )
   `add_new_ressl_cmd(this, create )
   `add_new_ressl_cmd(this, set )
   `add_new_ressl_cmd(this, add )
   `add_new_ressl_cmd(this, move )
    //`add_new_ressl_cmd(  this, copy )

   `add_new_ressl_cmd(this, attach )

   `add_new_ressl_cmd(this, start )
   `add_new_ressl_cmd(this, randomize )
   
   `add_new_ressl_cmd(this, load )
   `add_new_ressl_cmd(this, save )
    
   `add_new_ressl_cmd(this, quit )

   // Grab the default factory
   this.factory_ = uvm_factory::get();
   
   // ... etc.
endfunction : new

function boolean ressl::add_to_sequence_registry( 
   input string seq_name,
   input uvm_sequence_base seq
);
   seq_info my_seq_info;
   
   // ensure that it doesn't already exist. in the registry.
   if (this.m_sequence_registry.exists(seq_name)) begin
      this.stderr($sformatf("seq %s already exists.", seq_name),
                  .help_msg("You need to supply a unique name."));
      // todo: add the name of all the sequence names already in the associated
      // array appended to the help_msg.
      return( FALSE );
   end 

   my_seq_info = seq_info::type_id::create(seq_name);
   my_seq_info.set_seq(seq);
   // By default we attach the default_seqr, so it's always set to something.
   my_seq_info.set_seqr(this.m_default_seqr);
       
   this.m_sequence_registry.add(seq_name, my_seq_info);
   this.stdout( $sformatf("seq %s (type=%s) added.\n",
                          seq.get_name(),
                          seq.get_name()) );
   return( TRUE );
endfunction : add_to_sequence_registry

function boolean ressl::add_to_sequencer_registry( 
   input string             seqr_name,
   input uvm_sequencer_base seqr
);
   // ensure that it doesn't already exist. in the registry.
   if (this.m_sequencer_registry.exists(seqr_name)) begin
      this.stderr($sformatf("seqr %s already exists.", seqr_name),
                  .help_msg("You need to supply a unique name."));
      // todo: add the name of all the sequencer names already in the associated
      // array appended to the help_msg.
      return( FALSE );
   end 

   this.m_sequencer_registry[seqr_name] = seqr;
   this.stdout( $sformatf("seqr %s (type=%s) added.\n",
                          seqr.get_name(),
                          seqr.get_type_name()) );
   return( TRUE );
endfunction : add_to_sequencer_registry

function void ressl::set_default_seqr(
   input string seqr_name
);
   // The specified seqr has to exist in the repository first. 
   if (!this.m_sequencer_registry.exists(seqr_name)) begin
      this.stderr($sformatf("seqr %s is not in the sequencer repository.", seqr_name),
                  .help_msg("You need to add it to teh sequencer repository and then set the default."));
      // todo: add the name of all the sequencer names already in the associated
      // array appended to the help_msg.
   end
   else begin
      this.m_default_seqr = this.m_sequencer_registry[seqr_name];
   end
endfunction : set_default_seqr


`define STDIN   32'h8000_0000
`define STDOUT  32'h8000_0001
`define STDERR  32'h8000_0002 

`define CHAR_LF 'h0a
`define CHAR_SPACE 'h20
`define CHAR_ESC 'd27

`define CTRL 'h1d
`define UP_ARROW 'h57
`define DOWN_ARROW 'h62
   

// Function: get_user_response
//   A crude command line interface.  Provides a minimal amount of
//   history for recall.
function string ressl::get_user_response( 
);
   string           user_response;
   byte             c;
   string           c_str;
   int unsigned     cmd_ix;

   user_response = "";
   cmd_ix = 0;

   user_response = ressl_dpi_imported_readline(this.current_sequence_name_);
   return( user_response );

`ifdef RESSL_CRUDE_SV_CMDLINE
   
   this.stdout( $sformatf("[%s] %s[1m >>> ", this.current_sequence_name_, `CHAR_ESC));
   this.stdout( $sformatf("%s[0m", `CHAR_ESC));
   
   do begin
      c = $fgetc(`STDIN);

`ifdef BJM_CRUDE_HISTORY_IGNORED      
      if (c == `CTRL) begin
         $display("control pressed.");
         c = $fgetc(`STDIN);
         if ( (c == "p") ||(c == "P") ) begin
              c = `UP_ARROW;
         end
         if ( (c == "n") ||(c == "N") ) begin
              c = `DOWN_ARROW;
         end
      end
      
      if (c == `UP_ARROW) begin
         if (cmd_ix > 0) begin
            this.stdout( $sformatf("[%s] %s[1m >>> %s", this.current_sequence_name_,`CHAR_ESC, this.ressl_cmd_history[cmd_ix]));
            this.stdout( $sformatf("%s[0m", `CHAR_ESC));
         end
      end
      if (c == `DOWN_ARROW) begin
         if (cmd_ix < this.ressl_cmd_history.size()) begin
            this.stdout( $sformatf("[%s] %s[1m >>> %s", this.current_sequence_name_,`CHAR_ESC, this.ressl_cmd_history[cmd_ix]));
            this.stdout( $sformatf("%s[0m", `CHAR_ESC));
         end
      end
`endif //  `ifdef BJM_CRUDE_HISTORY_IGNORED
      
      if (c != `CHAR_LF) begin
         user_response = { user_response, string'(c) };
      end
   end while (c != `CHAR_LF);
   return( user_response );
`endif // RESSL_CRUDE_SV_CMDLINE
   
endfunction : get_user_response

// Task: interpret
//    Accepts the ~input_command~, tokenizes and then executes valid commands.
//    Invalid commands are ignored and result in error displayed.
// 
//    Returns the verb and arguments to allow further processing.
task ressl::interpret(
   input string    input_command,
   output string   verb,
   output string   arguments[$],
   output boolean  is_command_valid                   
);
   if ( this.tokenize(input_command, verb, arguments) == FALSE ) begin
      this.stderr(.err_msg($sformatf("Uh... what?!  Don't understand the command '%s'.\n", verb)),
                  .help_msg("Here are the list of commands you can use...\n"));
      is_command_valid = FALSE;
   end
   else begin
      // Add it to the history queue.
      this.ressl_cmd_history.push_front(input_command);

      this.execute( verb, arguments, is_command_valid );
   end
endtask : interpret

// Task: execute
//   Searches the verb list for the associated command then executes it.
//  Any error handling for the command is handled in the associated ressl_cmd.
task ressl::execute(
   input string  verb,
   input string  arguments[$],
   ref   boolean is_command_valid                 
);
   if (this.verbs_.exists(verb)) 
   begin
      ressl_cmd cmd;
      
      // Call the command pattern associated with this verb.  We pass it our
      // object so that it can manipulate the sequence registry.
      // see xample below
      // cmd = this.verb[my_verb];
      // cmd.execute(this, my_arguments);

      // Assume the command will work.  The verb itself will reset this value to
      // FALSE if it does not succeed. 
      is_command_valid = TRUE;
      
      this.verbs_[verb].execute( arguments, is_command_valid);

      // If the command ran sucessfully AND it should be saved into
      // the save/load history buffer, then do so.
      if (    (is_command_valid == TRUE)
           && (this.verbs_[verb].is_saveable() == TRUE)) begin
         svlib_pkg::Str all_args = svlib_pkg::Str::create(" ");
         this.save_load_history.push_back( { verb, " ", all_args.sjoin(arguments) } );
         // DEBUG foreach (this.save_load_history[i]) begin
         // DEBUG    $display("BJM: save_load_history[%0d] = %s",
         // DEBUG             i, this.save_load_history[i]);
         // DEBUG end
      end
   end
   else begin
      $display("Unrecognized command: %s", verb);
      is_command_valid = FALSE;
   end
endtask : execute

// Task: go
//    Starts the RESSL interpreter, gathers the user response and then runs each
//    verb
task ressl::go();

   string input_command;
   string my_verb;
   string my_arguments[$];
   boolean my_cmd_valid;
   
   this.stdout("SV-RESSL: Pinning Sequences to the Mat since 2014...\n");

   void'(ressl_dpi_imported_setup_readline());
   
   do begin
      this.interpret(.input_command( this.get_user_response() ),
                     .verb( my_verb ),
                     .arguments( my_arguments ),
                     .is_command_valid( my_cmd_valid ));

   end while ( my_verb != "quit"  );
endtask : go

function boolean ressl::tokenize(
   input string  input_command,
   output string verb,
   output string arguments[$]
);
   svlib_pkg::Str command = svlib_pkg::Str::create(input_command);
   svlib_private_base_pkg::qs  split_q       = command.split(" ");

   // NOTE: When you supply an empty string, string split returns exactly 1
   // element in the queue containing an empty string.   So the verb is
   // guaranteed to always be there, but if it's empty so are the arguments.
   verb      = split_q[0];
   if (verb != "") begin
      if (split_q.size() > 1) begin
         arguments = split_q[1:$];
      end
   
      return TRUE;
   end
   
   return FALSE;  // clearly kind of useless to supply an empty string.
endfunction : tokenize
   
function void ressl::stderr( 
   input string err_msg, 
   input string help_msg =""
);
   // err_msg = {err_msg,"\n"};
   help_msg = {help_msg,"\n"};
   $fwrite(`STDERR, $sformatf("%s[1m%s[31m ERR:%s[0m%s%s\n",
                              `CHAR_ESC, `CHAR_ESC, `CHAR_ESC, err_msg, help_msg ) );
//   $fwrite(`STDERR, { err_msg, help_msg } );
endfunction : stderr

function void ressl::stdout( 
   input string out_msg 
);
   // out_msg = {out_msg,"\n"};
   $fwrite(`STDOUT, out_msg );
endfunction : stdout



