//=============================================================================
//  @brief  types and classes for RESSL (Read-Evaluate-StartSequence-Loop)
//  @author Bryan Morris, Jeff McNeal, Verilab (www.verilab.com)
// =============================================================================
//
//                      svlib SystemVerilab Utilities Library
//
// @File: svlib_pkg_Ressl.svh
//
// Copyright 2015 Verilab, Inc.
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
//=============================================================================

class ressl extends uvm_object;
   // Holds a registry of handles to sequences indexed by name.  Where a name
   // can hold 1..N sequences.  Actually, we're holding a handle to a seq_info
   // which *contains* a seq.
   // Params: KEY    T
   registry#(string, seq_info) m_sequence_registry;

   // Hold a list of created sequencers so that we can attach a sequencer to
   // any sequence.
   uvm_sequencer_base m_default_seqr;
   uvm_sequencer_base m_sequencer_registry[string];

   // local uvm_sequence_base current_sequence_name_[$];
   local string current_sequence_name_;
   // local uvm_sequence_base current_recording_seq_[$];

   // Using the Command Pattern to hold each verb that the interpreter accepts.
   local ressl_cmd verbs_[string];

   // For the command-line history and save-load.  Add a history
   string ressl_cmd_history[$];
   
   // *Successful* command history and save-load.
   boolean is_cmd_ok;
   string save_load_history[$];

   // UVM Factory
   uvm_factory factory_;
   
   `uvm_object_utils_begin(ressl)
   `uvm_object_utils_end
   
   //
   extern function new(string name = "");
   extern task go();

   extern function void stderr( input string err_msg, input string help_msg = "" );
   extern function void stdout( input string out_msg );

   extern function boolean add_to_sequence_registry(
      input string            seq_name,                                              
      input uvm_sequence_base seq
   );
   extern function boolean add_to_sequencer_registry(
      input string             seqr_name,                                              
      input uvm_sequencer_base seqr
   );
   extern function void set_default_seqr(
      input string seqr_name
   );
                                      

   extern function string get_user_response();
   extern function boolean tokenize(
      input string    input_command,
      output string   verb,
      output string   arguments[$]
   );
   extern task interpret(
      input string    input_command,
      output string   verb,
      output string   arguments[$],
      output boolean  is_command_valid
   );
   extern task execute(
      input string   verb,
      input string   arguments[$],
      ref   boolean  is_command_valid                 
   );
                                
   function string get_current_seq_name();
      return this.current_sequence_name_;
   endfunction : get_current_seq_name
   
   function void set_current_sequence( input string seq_name );
      // TODO: Clarify this function's purpose.
      // if (seq_name != "") begin
      //    if (this.current_sequence_name_.get_name() != "") begin
      //       this.stdout($sformatf("Renaming current seqeunce from %s to %s",
      //                             this.current_sequence_name_.get_name(),
      //                             seq_name));
      //    end

      this.current_sequence_name_ = seq_name;
      // end
      // else begin
      //    this.stderr(.err_msg("I assumed you didn't want to name the current sequence as a empty string.  I ignored it for you."),
      //                .help_msg("...try again"));
      // end
   endfunction : set_current_sequence

   function void add_to_current_sequence( 
      input uvm_sequence_base add_sequence,
      input uvm_sequencer_base add_sequencer = null
    );

      seq_info my_seq_info;
       
      if (add_sequence != null) begin
         my_seq_info = seq_info::type_id::create(add_sequence.get_name());
         my_seq_info.set_seq(add_sequence);
         // Attach the sequencer they've specified, or use the default. 
         if (add_sequencer != null) my_seq_info.set_seqr(add_sequencer);
         else                       my_seq_info.set_seqr(this.m_default_seqr);
          
         this.m_sequence_registry.add(current_sequence_name_, my_seq_info);
      end
      else begin
         this.stderr("Attempting to set a null sequence. Ignored.");
      end
   endfunction : add_to_current_sequence

   // Function: get_seq
   //    Returns the handle to the <uvm_sequence_base> specified by the supplied
   //    <seq_index>.  Returns null if that is an invalid index.
   function uvm_sequence_base get_seq( input int unsigned seq_index );
      seq_info my_seq_info;
      
      if (seq_index < this.m_sequence_registry.num(current_sequence_name_)) begin
         if (this.m_sequence_registry.get_Nth(current_sequence_name_, seq_index, my_seq_info) == 0) begin
            // hmmm..couldn'tfind that sequence??  Not sure how that's possible.
            this.stderr( .err_msg( $sformatf("This is embarassing. I couldn't find sequence number %0d and it should be there. ", 
                                             seq_index) ),
                         .help_msg("Please raise a bug against us.  That's just not right."));
            get_seq = null;
         end
         else begin
            get_seq = my_seq_info.get_seq();
         end
      end
      else begin
         this.stderr( .err_msg( $sformatf("Uhh...sequence number %0d is not there. Try between 0 and %0d", 
                                          seq_index, 
                                          this.m_sequence_registry.num(current_sequence_name_))),
                      .help_msg("try again"));
         get_seq = null;
      end
   endfunction : get_seq

   function string help(input string verb);
      if (this.verbs_.exists(verb)) begin
         help = this.verbs_[verb].help();
         help = { help, "\n" };
      end 
      else begin
         this.stderr(.err_msg($sformatf("Ummm.. %s? no such verb.  Here's what's available.",verb)),
                     .help_msg(this.help_all()));
      end
   endfunction : help
   
   function string help_all();
      help_all = "\n--- RESSL Help --- \n";
      
      foreach (this.verbs_[key]) begin
         // aggregate the help from each command.
         help_all = { help_all, "\t", this.verbs_[key].help(), "\n" };
      end

      help_all = { help_all, "\n" };
   endfunction : help_all

   // add_new_ressl_cmd: Allows the end-user to create a new ressl_cmd and extend the
   // command interface.
   virtual function void add_new_ressl_cmd(
      input string     command_name,
      input ressl_cmd  command_handle
   );
      if (this.verbs_.exists(command_name)) begin
         this.stderr(.err_msg($sformatf("Attempting to add a new command (%s) that is already reserved.",command_name)),
                     .help_msg("Try using a different name for that command."));
      end
      else begin
         this.verbs_[command_name] = command_handle;
      end
   endfunction : add_new_ressl_cmd
                                
endclass : ressl

// ============================================================================
/////////////////// IMPLEMENTATIONS OF EXTERN CLASS METHODS ///////////////////

`include "ressl_impl_Ressl.svh"
