//=============================================================================
//  @brief  types and classes for the generic registry class.
//  @author Bryan Morris, Jeff McNeal, Verilab (www.verilab.com)
// =============================================================================
//
//                      svlib SystemVerilab Utilities Library
//
// @File: ressl_pkg_registry.svh
//
// Copyright 2015 Verilab, Inc.
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
//=============================================================================

// Class: registry
//   A generic class to create a simple registry, or a structure to hold q queue
//   of items each queue is associated with a key.
// 
//   In ASCII art the structure looks like (when the key is a string):
//
//   "name1" -> | item1a | item1b | item1c |
//   "name2" -> | item2a | item2b |
//
//   The user can either use this class directly (supplying a type of item and
//   type of key -- where the defaults for the item is uvm_sequence_base and the
//   key is a string.
//
//   The class also provides functions to search for an item based on name, and
//   add and delete items from the registry.
//
// Implementation:
// 
//   Currently, we use an associative array (keyed by the sequence name), to
//   hold the handle to the associated sequence.  This parameterized class
//   allows us to create other registries with different objects e.g., a
//   sequencer registry that can be search using the associative array.
//
//   NOTE: UVM-gurus: This is close-enough but not quite a uvm_pool, but I
//   (understandably) couldn't seem to coerce the T for ~uvm_pool~ to be a
//   queue.  So I created my own
//
class registry#(type KEY = string, type T = seq_info) extends uvm_object;
   local T registry_[KEY][$];

   // Used for iterating thru the q.
   // NOTE: That we keep track of each of the keys i.e., so you can be iterating
   //       thru any of the queues simultaneously.   Of course, we're assuming
   //       you're keeping track of which queue you're working on ;-)
   local int unsigned current_q_index_[KEY];

   // keys: really used for iterating thru the registry. e.g., 
   //   registry.keys(my_keys)
   //   foreach (my_keys[k]) 
   //      ... do something interesting here for each queue of T items in repository.
   local KEY keys_[$];
   
   `uvm_object_param_utils(registry)

   function new(string name = "registry");
      super.new(name);
   endfunction : new

   // Function: add
   //   Adds the supplied item_type into the registry, keyed by key.
   //   If key does not exist, a new queue of Ts is created and the
   //   item_type added to the front.
   //   If key already exists in the registry, the item_type is added to
   //   the back of the queue. 
   virtual function void add(
      input KEY key, 
      input T   item
   );
      this.reserve(key);
      this.registry_[key].push_back(item);
   endfunction : add

   // Function: reserve
   //   Reserve a key in the repository.  When you don't necessarily want to
   //   to hold a queue just yet.  Returns a 1 if the key is reserved or 0
   //   otherwise i.e., most likely coz it already exists. 
   virtual function int reserve(
      input KEY key
   );
       reserve = this.exists(key);
       if (reserve == 0) begin
          // doesn't exist. 
          this.keys_.push_back(key);
       end
   endfunction : reserve
   
   // Function: delete
   //   Remove the key AND the associated queue.  I hope you know what
   //   you're doing as there are no training wheels on this function. It
   //   deletes the queue and key without warnings of any kind. 
   virtual function void delete(
      input KEY key
   );
   int     key_indices[$];

      if (this.exists(key)) begin
         this.registry_[key].delete();
         key_indices = this.keys_.find_index(k) with (k == key);
         // should only be one...
         this.keys_.delete(key_indices[0]); 
      end
   endfunction : delete                          

   // Function: exists
   //   Does the key exist? 1 if yes; 0 no.
   virtual function exists(
      input KEY key
   );
      exists = this.registry_.exists(key);
   endfunction : exists

   // Function: keys
   //   Returns a queue with all the keys associated with this registry.
   virtual function void keys(
      ref KEY keys[$]
   );
      keys = this.keys_;
   endfunction : keys
   
   // Function: num
   //   The number of elements in the queue pointed to by the key.
   //   It returns 0 if the key is unknown.
   virtual function int num(
      input KEY key
   );
     if (this.exists(key)) num = this.registry_[key].size();
     else                  num = 0;
   endfunction : num

   // Function: get
   //   Returns a queue pointed to by the key (and it returns a 1), otherwise it
   //   returns a 0 and the q is unchanged.
   virtual function int get(
      input KEY    key,
      ref   T      q[$]
   );
      if (this.exists(key)) begin
         q = this.registry_[key];
         return(1);
      end
      return(0);
   endfunction : get   
                  
   // Function: get_Nth
   //   Returns the Nth item in the queue pointed by the key.  
   //   Recall the 0th element is the first item in the queue.
   //   A 1 indicates the returned item is valid; 0 means the item is
   //   unchanged. 
   virtual function int get_Nth(
      input KEY          key,                          
      input int unsigned Nth,
      ref   T            item                           
   );
      if (this.exists(key)) begin
         if (this.registry_[key].size() > Nth) begin
            item = this.registry_[key][Nth];
            return(1);
         end
      end
      return(0);  // item is unchnaged.
   endfunction : get_Nth

   // Function: first
   //   Returns the first item in the queue pointed to by the key.
   //   The function returns 1 if we found the first; and 0 otherwise and 
   //   a null is returned if the key supplied is unknown (or the queue it's
   //   pointing is empty.)  
   virtual function int first(
      input KEY key,
      ref   T   item                            
   );
      // if we can find the element with get_Nth, then the key exists, and the q
      // is not empty, so we can store the current iteration state.
      if (this.get_Nth(key, 0, item) == 0) begin
         this.current_q_index_[key] = 0;
         return(1);
      end
      return(0);
   
   endfunction : first

   // Function: next
   //   Returns the next item in the queue pointed to by the key.  The
   //   The function returns 1 if item is in the queue and is NOT the last item,
   //   and 0 if this is the last item in the queue.  The assumption is that
   //   you've called first already to start the iteration.
   // 
   //   A null is returned if the key supplied is unknown (or you've asked for
   //   the next item beyond the end of the queue. 
   virtual function int next(
      input KEY key,
      ref   T   item                            
   );
      if (this.exists(key)) begin
         if (this.get_Nth(key, this.current_q_index_[key], item) == 0) begin
            // Is this the last item in the queue?
            if (this.current_q_index_[key] <= this.registry_[key].size()-1) begin
               return(1);  // there are still more elements
            end
         end
      end
      return(0);  // no more elements in the queue
   endfunction : next

   // Function: move
   //   Move one element from_index to_index in the queue indicated by thekey
   virtual function int move(
      input KEY      key,
      input int unsigned from_index,
      input int unsigned to_index
   );
      if (this.exists(key)) begin
         this.registry_[key].insert(to_index,
                                    this.registry_[key][from_index]);
         foreach (this.registry_[key][i]) begin
            $display("after insert: %0d : %s",
                     i, this.registry_[key][i].get_name());
         end
         
         // moving up or down?
         if (from_index <= to_index) begin
            this.registry_[key].delete(from_index);
            $display("delete @ %0d", from_index);
         end
         else  begin
            this.registry_[key].delete(from_index+1);
            $display("delete @ %0d", from_index+1);
         end
         foreach (this.registry_[key][i]) begin
            $display("after delete: %0d : %s",
                     i, this.registry_[key][i].get_name());
         end
      end
   endfunction : move                       
endclass : registry



  






